@@ -42,7 +42,7 @@ int TestPathMakePath(int argc, char* argv[])
 	}
 
 	printf("Creating path %s\n", path);
-	success = PathMakePathA(path, NULL);
+	success = winpr_PathMakePath(path, NULL);
 
 	if (!success)
 	{
@@ -51,7 +51,7 @@ int TestPathMakePath(int argc, char* argv[])
 		return -1;
 	}
 
-	success = PathFileExistsA(path);
+	success = winpr_PathFileExists(path);
 
 	if (!success)
 	{
@@ -62,9 +62,9 @@ int TestPathMakePath(int argc, char* argv[])
 
 	while (strlen(path) > baseLen)
 	{
-		if (!RemoveDirectoryA(path))
+		if (!winpr_RemoveDirectory(path))
 		{
-			fprintf(stderr, "RemoveDirectoryA %s failed!\n", path);
+			fprintf(stderr, "winpr_RemoveDirectory %s failed!\n", path);
 			free(path);
 			return -1;
 		}
