@@ -4,6 +4,8 @@
  *
  * Copyright 2012 Marc-Andre Moreau <marcandre.moreau@gmail.com>
  * Copyright 2015 Hewlett-Packard Development Company, L.P.
+ * Copyright 2021 David Fort <contact@hardening-consulting.com>
+ *
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -22,7 +24,7 @@
 #include "config.h"
 #endif
 
-#include <assert.h>
+#include <winpr/assert.h>
 
 #include <winpr/handle.h>
 
@@ -89,12 +91,17 @@
 #include <winpr/collections.h>
 
 #include "thread.h"
+#include "apc.h"
 
 #include "../handle/handle.h"
 #include "../log.h"
 #define TAG WINPR_TAG("thread")
 
+static WINPR_THREAD mainThread;
+
+#if defined(WITH_THREAD_LIST)
 static wListDictionary* thread_list = NULL;
+#endif
 
 static BOOL ThreadCloseHandle(HANDLE handle);
 static void cleanup_handle(void* obj);
@@ -119,17 +126,193 @@ static int ThreadGetFd(HANDLE handle)
 	if (!ThreadIsHandled(handle))
 		return -1;
 
-	return pThread->pipe_fd[0];
+	return pThread->event.fds[0];
+}
+
+#define run_mutex_init(fkt, mux, arg) run_mutex_init_(fkt, #fkt, mux, arg)
+static BOOL run_mutex_init_(int (*fkt)(pthread_mutex_t*, const pthread_mutexattr_t*),
+                            const char* name, pthread_mutex_t* mutex,
+                            const pthread_mutexattr_t* mutexattr)
+{
+	int rc;
+
+	WINPR_ASSERT(fkt);
+	WINPR_ASSERT(mutex);
+
+	rc = fkt(mutex, mutexattr);
+	if (rc != 0)
+	{
+		WLog_WARN(TAG, "[%s] failed with [%s]", name, strerror(rc));
+	}
+	return rc == 0;
+}
+
+#define run_mutex_fkt(fkt, mux) run_mutex_fkt_(fkt, #fkt, mux)
+static BOOL run_mutex_fkt_(int (*fkt)(pthread_mutex_t* mux), const char* name,
+                           pthread_mutex_t* mutex)
+{
+	int rc;
+
+	WINPR_ASSERT(fkt);
+	WINPR_ASSERT(mutex);
+
+	rc = fkt(mutex);
+	if (rc != 0)
+	{
+		WLog_WARN(TAG, "[%s] failed with [%s]", name, strerror(rc));
+	}
+	return rc == 0;
+}
+
+#define run_cond_init(fkt, cond, arg) run_cond_init_(fkt, #fkt, cond, arg)
+static BOOL run_cond_init_(int (*fkt)(pthread_cond_t*, const pthread_condattr_t*), const char* name,
+                           pthread_cond_t* condition, const pthread_condattr_t* conditionattr)
+{
+	int rc;
+
+	WINPR_ASSERT(fkt);
+	WINPR_ASSERT(condition);
+
+	rc = fkt(condition, conditionattr);
+	if (rc != 0)
+	{
+		WLog_WARN(TAG, "[%s] failed with [%s]", name, strerror(rc));
+	}
+	return rc == 0;
+}
+
+#define run_cond_fkt(fkt, cond) run_cond_fkt_(fkt, #fkt, cond)
+static BOOL run_cond_fkt_(int (*fkt)(pthread_cond_t* mux), const char* name,
+                          pthread_cond_t* condition)
+{
+	int rc;
+
+	WINPR_ASSERT(fkt);
+	WINPR_ASSERT(condition);
+
+	rc = fkt(condition);
+	if (rc != 0)
+	{
+		WLog_WARN(TAG, "[%s] failed with [%s]", name, strerror(rc));
+	}
+	return rc == 0;
+}
+
+static int pthread_mutex_checked_unlock(pthread_mutex_t* mutex)
+{
+	WINPR_ASSERT(mutex);
+	WINPR_ASSERT(pthread_mutex_trylock(mutex) == EBUSY);
+	return pthread_mutex_unlock(mutex);
+}
+
+static BOOL mux_condition_bundle_init(mux_condition_bundle* bundle)
+{
+	WINPR_ASSERT(bundle);
+
+	bundle->val = FALSE;
+	if (!run_mutex_init(pthread_mutex_init, &bundle->mux, NULL))
+		return FALSE;
+
+	if (!run_cond_init(pthread_cond_init, &bundle->cond, NULL))
+		return FALSE;
+	return TRUE;
+}
+
+static void mux_condition_bundle_uninit(mux_condition_bundle* bundle)
+{
+	mux_condition_bundle empty = { 0 };
+
+	WINPR_ASSERT(bundle);
+
+	run_cond_fkt(pthread_cond_destroy, &bundle->cond);
+	run_mutex_fkt(pthread_mutex_destroy, &bundle->mux);
+	*bundle = empty;
+}
+
+static BOOL mux_condition_bundle_signal(mux_condition_bundle* bundle)
+{
+	BOOL rc = TRUE;
+	WINPR_ASSERT(bundle);
+
+	if (!run_mutex_fkt(pthread_mutex_lock, &bundle->mux))
+		return FALSE;
+	bundle->val = TRUE;
+	if (!run_cond_fkt(pthread_cond_signal, &bundle->cond))
+		rc = FALSE;
+	if (!run_mutex_fkt(pthread_mutex_checked_unlock, &bundle->mux))
+		rc = FALSE;
+	return rc;
+}
+
+static BOOL mux_condition_bundle_lock(mux_condition_bundle* bundle)
+{
+	WINPR_ASSERT(bundle);
+	return run_mutex_fkt(pthread_mutex_lock, &bundle->mux);
+}
+
+static BOOL mux_condition_bundle_unlock(mux_condition_bundle* bundle)
+{
+	WINPR_ASSERT(bundle);
+	return run_mutex_fkt(pthread_mutex_checked_unlock, &bundle->mux);
+}
+
+static BOOL mux_condition_bundle_wait(mux_condition_bundle* bundle, const char* name)
+{
+	BOOL rc = FALSE;
+
+	WINPR_ASSERT(bundle);
+	WINPR_ASSERT(name);
+	WINPR_ASSERT(pthread_mutex_trylock(&bundle->mux) == EBUSY);
+
+	while (!bundle->val)
+	{
+		int r = pthread_cond_wait(&bundle->cond, &bundle->mux);
+		if (r != 0)
+		{
+			WLog_ERR(TAG, "failed to wait for %s [%s]", name, strerror(r));
+			switch (r)
+			{
+				case ENOTRECOVERABLE:
+				case EPERM:
+				case ETIMEDOUT:
+				case EINVAL:
+					goto fail;
+
+				default:
+					break;
+			}
+		}
+	}
+
+	rc = bundle->val;
+
+fail:
+	return rc;
+}
+
+static BOOL signal_thread_ready(WINPR_THREAD* thread)
+{
+	WINPR_ASSERT(thread);
+
+	return mux_condition_bundle_signal(&thread->isCreated);
+}
+
+static BOOL signal_thread_is_running(WINPR_THREAD* thread)
+{
+	WINPR_ASSERT(thread);
+
+	return mux_condition_bundle_signal(&thread->isRunning);
 }
 
 static DWORD ThreadCleanupHandle(HANDLE handle)
 {
+	DWORD status = WAIT_FAILED;
 	WINPR_THREAD* thread = (WINPR_THREAD*)handle;
 
 	if (!ThreadIsHandled(handle))
 		return WAIT_FAILED;
 
-	if (pthread_mutex_lock(&thread->mutex))
+	if (!run_mutex_fkt(pthread_mutex_lock, &thread->mutex))
 		return WAIT_FAILED;
 
 	if (!thread->joined)
@@ -140,17 +323,19 @@ static DWORD ThreadCleanupHandle(HANDLE handle)
 		if (status != 0)
 		{
 			WLog_ERR(TAG, "pthread_join failure: [%d] %s", status, strerror(status));
-			pthread_mutex_unlock(&thread->mutex);
-			return WAIT_FAILED;
+			goto fail;
 		}
 		else
 			thread->joined = TRUE;
 	}
 
-	if (pthread_mutex_unlock(&thread->mutex))
+	status = WAIT_OBJECT_0;
+
+fail:
+	if (!run_mutex_fkt(pthread_mutex_checked_unlock, &thread->mutex))
 		return WAIT_FAILED;
 
-	return WAIT_OBJECT_0;
+	return status;
 }
 
 static HANDLE_OPS ops = { ThreadIsHandled,
@@ -214,7 +399,8 @@ static void dump_thread(WINPR_THREAD* thread)
 
 		free(msg);
 	}
-
+#else
+	WINPR_UNUSED(thread);
 #endif
 }
 
@@ -224,60 +410,15 @@ static void dump_thread(WINPR_THREAD* thread)
  */
 static BOOL set_event(WINPR_THREAD* thread)
 {
-	int length;
-	BOOL status = FALSE;
-#ifdef HAVE_SYS_EVENTFD_H
-	eventfd_t val = 1;
-
-	do
-	{
-		length = eventfd_write(thread->pipe_fd[0], val);
-	} while ((length < 0) && (errno == EINTR));
-
-	status = (length == 0) ? TRUE : FALSE;
-#else
-
-	if (WaitForSingleObject(thread, 0) != WAIT_OBJECT_0)
-	{
-		length = write(thread->pipe_fd[1], "-", 1);
-
-		if (length == 1)
-			status = TRUE;
-	}
-	else
-	{
-		status = TRUE;
-	}
-
-#endif
-	return status;
+	return winpr_event_set(&thread->event);
 }
 
 static BOOL reset_event(WINPR_THREAD* thread)
 {
-	int length;
-	BOOL status = FALSE;
-#ifdef HAVE_SYS_EVENTFD_H
-	eventfd_t value;
-
-	do
-	{
-		length = eventfd_read(thread->pipe_fd[0], &value);
-	} while ((length < 0) && (errno == EINTR));
-
-	if ((length > 0) && (!status))
-		status = TRUE;
-
-#else
-	length = read(thread->pipe_fd[0], &length, 1);
-
-	if ((length == 1) && (!status))
-		status = TRUE;
-
-#endif
-	return status;
+	return winpr_event_reset(&thread->event);
 }
 
+#if defined(WITH_THREAD_LIST)
 static BOOL thread_compare(const void* a, const void* b)
 {
 	const pthread_t* p1 = a;
@@ -285,6 +426,76 @@ static BOOL thread_compare(const void* a, const void* b)
 	BOOL rc = pthread_equal(*p1, *p2);
 	return rc;
 }
+#endif
+
+static INIT_ONCE threads_InitOnce = INIT_ONCE_STATIC_INIT;
+static pthread_t mainThreadId;
+static DWORD currentThreadTlsIndex = TLS_OUT_OF_INDEXES;
+
+static BOOL initializeThreads(PINIT_ONCE InitOnce, PVOID Parameter, PVOID* Context)
+{
+	if (!apc_init(&mainThread.apc))
+	{
+		WLog_ERR(TAG, "failed to initialize APC");
+		goto out;
+	}
+
+	mainThread.Type = HANDLE_TYPE_THREAD;
+	mainThreadId = pthread_self();
+
+	currentThreadTlsIndex = TlsAlloc();
+	if (currentThreadTlsIndex == TLS_OUT_OF_INDEXES)
+	{
+		WLog_ERR(TAG, "Major bug, unable to allocate a TLS value for currentThread");
+	}
+
+#if defined(WITH_THREAD_LIST)
+	thread_list = ListDictionary_New(TRUE);
+
+	if (!thread_list)
+	{
+		WLog_ERR(TAG, "Couldn't create global thread list");
+		goto error_thread_list;
+	}
+
+	thread_list->objectKey.fnObjectEquals = thread_compare;
+#endif
+
+out:
+	return TRUE;
+}
+
+static BOOL signal_and_wait_for_ready(WINPR_THREAD* thread)
+{
+	BOOL res = FALSE;
+
+	WINPR_ASSERT(thread);
+
+	if (!mux_condition_bundle_lock(&thread->isRunning))
+		return FALSE;
+
+	if (!signal_thread_ready(thread))
+		goto fail;
+
+	if (!mux_condition_bundle_wait(&thread->isRunning, "threadIsRunning"))
+		goto fail;
+
+#if defined(WITH_THREAD_LIST)
+	if (!ListDictionary_Contains(thread_list, &thread->thread))
+	{
+		WLog_ERR(TAG, "Thread not in thread_list, startup failed!");
+		goto fail;
+	}
+#endif
+
+	res = TRUE;
+
+fail:
+	if (!mux_condition_bundle_unlock(&thread->isRunning))
+		return FALSE;
+
+	return res;
+}
 
 /* Thread launcher function responsible for registering
  * cleanup handlers and calling pthread_exit, if not done
@@ -301,39 +512,35 @@ static void* thread_launcher(void* arg)
 		goto exit;
 	}
 
+	if (!TlsSetValue(currentThreadTlsIndex, thread))
+	{
+		WLog_ERR(TAG, "thread %d, unable to set current thread value", pthread_self());
+		goto exit;
+	}
+
 	if (!(fkt = thread->lpStartAddress))
 	{
 		WLog_ERR(TAG, "Thread function argument is %p", (void*)fkt);
 		goto exit;
 	}
 
-	if (pthread_mutex_lock(&thread->threadIsReadyMutex))
+	if (!signal_and_wait_for_ready(thread))
 		goto exit;
 
-	if (!ListDictionary_Contains(thread_list, &thread->thread))
-	{
-		if (pthread_cond_wait(&thread->threadIsReady, &thread->threadIsReadyMutex) != 0)
-		{
-			WLog_ERR(TAG, "The thread could not be made ready");
-			pthread_mutex_unlock(&thread->threadIsReadyMutex);
-			goto exit;
-		}
-	}
-
-	if (pthread_mutex_unlock(&thread->threadIsReadyMutex))
-		goto exit;
-
-	assert(ListDictionary_Contains(thread_list, &thread->thread));
 	rc = fkt(thread->lpParameter);
 exit:
 
 	if (thread)
 	{
+		apc_cleanupThread(thread);
+
 		if (!thread->exited)
 			thread->dwExitCode = rc;
 
 		set_event(thread);
 
+		signal_thread_ready(thread);
+
 		if (thread->detached || !thread->started)
 			cleanup_handle(thread);
 	}
@@ -343,7 +550,14 @@ exit:
 
 static BOOL winpr_StartThread(WINPR_THREAD* thread)
 {
-	pthread_attr_t attr;
+	BOOL rc = FALSE;
+	BOOL locked = FALSE;
+	pthread_attr_t attr = { 0 };
+
+	if (!mux_condition_bundle_lock(&thread->isCreated))
+		return FALSE;
+	locked = TRUE;
+
 	pthread_attr_init(&attr);
 	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 
@@ -353,35 +567,44 @@ static BOOL winpr_StartThread(WINPR_THREAD* thread)
 	thread->started = TRUE;
 	reset_event(thread);
 
-	if (pthread_create(&thread->thread, &attr, thread_launcher, thread))
-		goto error;
-
-	if (pthread_mutex_lock(&thread->threadIsReadyMutex))
-		goto error;
-
+#if defined(WITH_THREAD_LIST)
 	if (!ListDictionary_Add(thread_list, &thread->thread, thread))
 	{
 		WLog_ERR(TAG, "failed to add the thread to the thread list");
-		pthread_mutex_unlock(&thread->threadIsReadyMutex);
 		goto error;
 	}
+#endif
 
-	if (pthread_cond_signal(&thread->threadIsReady) != 0)
+	if (pthread_create(&thread->thread, &attr, thread_launcher, thread))
+		goto error;
+
+	if (!mux_condition_bundle_wait(&thread->isCreated, "threadIsCreated"))
+		goto error;
+
+	locked = FALSE;
+	if (!mux_condition_bundle_unlock(&thread->isCreated))
+		goto error;
+
+	if (!signal_thread_is_running(thread))
 	{
 		WLog_ERR(TAG, "failed to signal the thread was ready");
-		pthread_mutex_unlock(&thread->threadIsReadyMutex);
 		goto error;
 	}
 
-	if (pthread_mutex_unlock(&thread->threadIsReadyMutex))
-		goto error;
+	rc = TRUE;
+error:
+	if (locked)
+	{
+		if (!mux_condition_bundle_unlock(&thread->isCreated))
+			rc = FALSE;
+	}
 
 	pthread_attr_destroy(&attr);
-	dump_thread(thread);
-	return TRUE;
-error:
-	pthread_attr_destroy(&attr);
-	return FALSE;
+
+	if (rc)
+		dump_thread(thread);
+
+	return rc;
 }
 
 HANDLE CreateThread(LPSECURITY_ATTRIBUTES lpThreadAttributes, SIZE_T dwStackSize,
@@ -389,8 +612,7 @@ HANDLE CreateThread(LPSECURITY_ATTRIBUTES lpThreadAttributes, SIZE_T dwStackSize
                     DWORD dwCreationFlags, LPDWORD lpThreadId)
 {
 	HANDLE handle;
-	WINPR_THREAD* thread;
-	thread = (WINPR_THREAD*)calloc(1, sizeof(WINPR_THREAD));
+	WINPR_THREAD* thread = (WINPR_THREAD*)calloc(1, sizeof(WINPR_THREAD));
 
 	if (!thread)
 		return NULL;
@@ -404,127 +626,70 @@ HANDLE CreateThread(LPSECURITY_ATTRIBUTES lpThreadAttributes, SIZE_T dwStackSize
 	thread->create_stack = winpr_backtrace(20);
 	dump_thread(thread);
 #endif
-	thread->pipe_fd[0] = -1;
-	thread->pipe_fd[1] = -1;
-#ifdef HAVE_SYS_EVENTFD_H
-	thread->pipe_fd[0] = eventfd(0, EFD_NONBLOCK);
 
-	if (thread->pipe_fd[0] < 0)
+	if (!winpr_event_init(&thread->event))
 	{
-		WLog_ERR(TAG, "failed to create thread pipe fd 0");
-		goto error_pipefd0;
+		WLog_ERR(TAG, "failed to create event");
+		goto fail;
 	}
 
-#else
-
-	if (pipe(thread->pipe_fd) < 0)
-	{
-		WLog_ERR(TAG, "failed to create thread pipe");
-		goto error_pipefd0;
-	}
-
-	{
-		int flags = fcntl(thread->pipe_fd[0], F_GETFL);
-		fcntl(thread->pipe_fd[0], F_SETFL, flags | O_NONBLOCK);
-	}
-
-#endif
-
-	if (pthread_mutex_init(&thread->mutex, 0) != 0)
+	if (!run_mutex_init(pthread_mutex_init, &thread->mutex, NULL))
 	{
 		WLog_ERR(TAG, "failed to initialize thread mutex");
-		goto error_mutex;
+		goto fail;
 	}
 
-	if (pthread_mutex_init(&thread->threadIsReadyMutex, NULL) != 0)
+	if (!apc_init(&thread->apc))
 	{
-		WLog_ERR(TAG, "failed to initialize a mutex for a condition variable");
-		goto error_thread_ready_mutex;
+		WLog_ERR(TAG, "failed to initialize APC");
+		goto fail;
 	}
 
-	if (pthread_cond_init(&thread->threadIsReady, NULL) != 0)
-	{
-		WLog_ERR(TAG, "failed to initialize a condition variable");
-		goto error_thread_ready;
-	}
+	if (!mux_condition_bundle_init(&thread->isCreated))
+		goto fail;
+	if (!mux_condition_bundle_init(&thread->isRunning))
+		goto fail;
 
 	WINPR_HANDLE_SET_TYPE_AND_MODE(thread, HANDLE_TYPE_THREAD, WINPR_FD_READ);
 	handle = (HANDLE)thread;
 
-	if (!thread_list)
-	{
-		thread_list = ListDictionary_New(TRUE);
-
-		if (!thread_list)
-		{
-			WLog_ERR(TAG, "Couldn't create global thread list");
-			goto error_thread_list;
-		}
-
-		thread_list->objectKey.fnObjectEquals = thread_compare;
-	}
+	InitOnceExecuteOnce(&threads_InitOnce, initializeThreads, NULL, NULL);
 
 	if (!(dwCreationFlags & CREATE_SUSPENDED))
 	{
 		if (!winpr_StartThread(thread))
-			goto error_thread_list;
+			goto fail;
 	}
 	else
 	{
 		if (!set_event(thread))
-			goto error_thread_list;
+			goto fail;
 	}
 
 	return handle;
-error_thread_list:
-	pthread_cond_destroy(&thread->threadIsReady);
-error_thread_ready:
-	pthread_mutex_destroy(&thread->threadIsReadyMutex);
-error_thread_ready_mutex:
-	pthread_mutex_destroy(&thread->mutex);
-error_mutex:
-
-	if (thread->pipe_fd[1] >= 0)
-		close(thread->pipe_fd[1]);
-
-	if (thread->pipe_fd[0] >= 0)
-		close(thread->pipe_fd[0]);
-
-error_pipefd0:
-	free(thread);
+fail:
+	cleanup_handle(thread);
 	return NULL;
 }
 
 void cleanup_handle(void* obj)
 {
-	int rc;
 	WINPR_THREAD* thread = (WINPR_THREAD*)obj;
-	rc = pthread_cond_destroy(&thread->threadIsReady);
+	if (!thread)
+		return;
 
-	if (rc)
-		WLog_ERR(TAG, "failed to destroy a condition variable [%d] %s (%d)", rc, strerror(errno),
-		         errno);
+	if (!apc_uninit(&thread->apc))
+		WLog_ERR(TAG, "failed to destroy APC");
 
-	rc = pthread_mutex_destroy(&thread->threadIsReadyMutex);
+	mux_condition_bundle_uninit(&thread->isCreated);
+	mux_condition_bundle_uninit(&thread->isRunning);
+	run_mutex_fkt(pthread_mutex_destroy, &thread->mutex);
 
-	if (rc)
-		WLog_ERR(TAG, "failed to destroy a condition variable mutex [%d] %s (%d)", rc,
-		         strerror(errno), errno);
-
-	rc = pthread_mutex_destroy(&thread->mutex);
-
-	if (rc)
-		WLog_ERR(TAG, "failed to destroy mutex [%d] %s (%d)", rc, strerror(errno), errno);
-
-	if (thread->pipe_fd[0] >= 0)
-		close(thread->pipe_fd[0]);
-
-	if (thread->pipe_fd[1] >= 0)
-		close(thread->pipe_fd[1]);
-
-	if (thread_list && ListDictionary_Contains(thread_list, &thread->thread))
-		ListDictionary_Remove(thread_list, &thread->thread);
+	winpr_event_uninit(&thread->event);
 
+#if defined(WITH_THREAD_LIST)
+	ListDictionary_Remove(thread_list, &thread->thread);
+#endif
 #if defined(WITH_DEBUG_THREADS)
 
 	if (thread->create_stack)
@@ -541,6 +706,7 @@ BOOL ThreadCloseHandle(HANDLE handle)
 {
 	WINPR_THREAD* thread = (WINPR_THREAD*)handle;
 
+#if defined(WITH_THREAD_LIST)
 	if (!thread_list)
 	{
 		WLog_ERR(TAG, "Thread list does not exist, check call!");
@@ -554,6 +720,7 @@ BOOL ThreadCloseHandle(HANDLE handle)
 	else
 	{
 		ListDictionary_Lock(thread_list);
+#endif
 		dump_thread(thread);
 
 		if ((thread->started) && (WaitForSingleObject(thread, 0) != WAIT_OBJECT_0))
@@ -567,14 +734,10 @@ BOOL ThreadCloseHandle(HANDLE handle)
 			cleanup_handle(thread);
 		}
 
+#if defined(WITH_THREAD_LIST)
 		ListDictionary_Unlock(thread_list);
-
-		if (ListDictionary_Count(thread_list) < 1)
-		{
-			ListDictionary_Free(thread_list);
-			thread_list = NULL;
-		}
 	}
+#endif
 
 	return TRUE;
 }
@@ -590,6 +753,7 @@ HANDLE CreateRemoteThread(HANDLE hProcess, LPSECURITY_ATTRIBUTES lpThreadAttribu
 
 VOID ExitThread(DWORD dwExitCode)
 {
+#if defined(WITH_THREAD_LIST)
 	DWORD rc;
 	pthread_t tid = pthread_self();
 
@@ -614,7 +778,7 @@ VOID ExitThread(DWORD dwExitCode)
 		WINPR_THREAD* thread;
 		ListDictionary_Lock(thread_list);
 		thread = ListDictionary_GetItemValue(thread_list, &tid);
-		assert(thread);
+		WINPR_ASSERT(thread);
 		thread->exited = TRUE;
 		thread->dwExitCode = dwExitCode;
 #if defined(WITH_DEBUG_THREADS)
@@ -629,6 +793,9 @@ VOID ExitThread(DWORD dwExitCode)
 
 		pthread_exit((void*)(size_t)rc);
 	}
+#else
+	WINPR_UNUSED(dwExitCode);
+#endif
 }
 
 BOOL GetExitCodeThread(HANDLE hThread, LPDWORD lpExitCode)
@@ -645,31 +812,28 @@ BOOL GetExitCodeThread(HANDLE hThread, LPDWORD lpExitCode)
 	return TRUE;
 }
 
-HANDLE _GetCurrentThread(VOID)
+WINPR_THREAD* winpr_GetCurrentThread(VOID)
 {
-	HANDLE hdl = NULL;
-	pthread_t tid = pthread_self();
+	WINPR_THREAD* ret;
 
-	if (!thread_list)
-	{
-		WLog_ERR(TAG, "function called without existing thread list!");
-#if defined(WITH_DEBUG_THREADS)
-		DumpThreadHandles();
-#endif
-	}
-	else if (!ListDictionary_Contains(thread_list, &tid))
+	InitOnceExecuteOnce(&threads_InitOnce, initializeThreads, NULL, NULL);
+	if (mainThreadId == pthread_self())
+		return (HANDLE)&mainThread;
+
+	ret = TlsGetValue(currentThreadTlsIndex);
+	if (!ret)
 	{
 		WLog_ERR(TAG, "function called, but no matching entry in thread list!");
 #if defined(WITH_DEBUG_THREADS)
 		DumpThreadHandles();
 #endif
 	}
-	else
-	{
-		hdl = ListDictionary_GetItemValue(thread_list, &tid);
-	}
+	return ret;
+}
 
-	return hdl;
+HANDLE _GetCurrentThread(VOID)
+{
+	return (HANDLE)winpr_GetCurrentThread();
 }
 
 DWORD GetCurrentThreadId(VOID)
@@ -681,6 +845,58 @@ DWORD GetCurrentThreadId(VOID)
 	return (DWORD)tid & 0xffffffffUL;
 }
 
+typedef struct
+{
+	WINPR_APC_ITEM apc;
+	PAPCFUNC completion;
+	ULONG_PTR completionArg;
+} UserApcItem;
+
+static void userAPC(LPVOID arg)
+{
+	UserApcItem* userApc = (UserApcItem*)arg;
+
+	userApc->completion(userApc->completionArg);
+
+	userApc->apc.markedForRemove = TRUE;
+}
+
+DWORD QueueUserAPC(PAPCFUNC pfnAPC, HANDLE hThread, ULONG_PTR dwData)
+{
+	ULONG Type;
+	WINPR_HANDLE* Object;
+	WINPR_APC_ITEM* apc;
+	UserApcItem* apcItem;
+
+	if (!pfnAPC)
+		return 1;
+
+	if (!winpr_Handle_GetInfo(hThread, &Type, &Object) || Object->Type != HANDLE_TYPE_THREAD)
+	{
+		WLog_ERR(TAG, "hThread is not a thread");
+		SetLastError(ERROR_INVALID_PARAMETER);
+		return (DWORD)0;
+	}
+
+	apcItem = calloc(1, sizeof(*apcItem));
+	if (!apcItem)
+	{
+		SetLastError(ERROR_INVALID_PARAMETER);
+		return (DWORD)0;
+	}
+
+	apc = &apcItem->apc;
+	apc->type = APC_TYPE_USER;
+	apc->markedForFree = TRUE;
+	apc->alwaysSignaled = TRUE;
+	apc->completion = userAPC;
+	apc->completionArgs = apc;
+	apcItem->completion = pfnAPC;
+	apcItem->completionArg = dwData;
+	apc_register(hThread, apc);
+	return 1;
+}
+
 DWORD ResumeThread(HANDLE hThread)
 {
 	ULONG Type;
@@ -692,21 +908,21 @@ DWORD ResumeThread(HANDLE hThread)
 
 	thread = (WINPR_THREAD*)Object;
 
-	if (pthread_mutex_lock(&thread->mutex))
+	if (!run_mutex_fkt(pthread_mutex_lock, &thread->mutex))
 		return (DWORD)-1;
 
 	if (!thread->started)
 	{
 		if (!winpr_StartThread(thread))
 		{
-			pthread_mutex_unlock(&thread->mutex);
+			run_mutex_fkt(pthread_mutex_checked_unlock, &thread->mutex);
 			return (DWORD)-1;
 		}
 	}
 	else
 		WLog_WARN(TAG, "Thread already started!");
 
-	if (pthread_mutex_unlock(&thread->mutex))
+	if (!run_mutex_fkt(pthread_mutex_checked_unlock, &thread->mutex))
 		return (DWORD)-1;
 
 	return 0;
@@ -744,7 +960,7 @@ BOOL TerminateThread(HANDLE hThread, DWORD dwExitCode)
 	thread->exited = TRUE;
 	thread->dwExitCode = dwExitCode;
 
-	if (pthread_mutex_lock(&thread->mutex))
+	if (!run_mutex_fkt(pthread_mutex_lock, &thread->mutex))
 		return FALSE;
 
 #ifndef ANDROID
@@ -753,16 +969,16 @@ BOOL TerminateThread(HANDLE hThread, DWORD dwExitCode)
 	WLog_ERR(TAG, "Function not supported on this platform!");
 #endif
 
-	if (pthread_mutex_unlock(&thread->mutex))
+	if (!run_mutex_fkt(pthread_mutex_checked_unlock, &thread->mutex))
 		return FALSE;
 
 	set_event(thread);
 	return TRUE;
 }
 
-#if defined(WITH_DEBUG_THREADS)
 VOID DumpThreadHandles(void)
 {
+#if defined(WITH_DEBUG_THREADS)
 	char** msg;
 	size_t used, i;
 	void* stack = winpr_backtrace(20);
@@ -778,6 +994,7 @@ VOID DumpThreadHandles(void)
 	winpr_backtrace_free(stack);
 	WLog_DBG(TAG, "---------------- Start Dumping thread handles -----------");
 
+#if defined(WITH_THREAD_LIST)
 	if (!thread_list)
 	{
 		WLog_DBG(TAG, "All threads properly shut down and disposed of.");
@@ -821,8 +1038,9 @@ VOID DumpThreadHandles(void)
 		free(keys);
 		ListDictionary_Unlock(thread_list);
 	}
+#endif
 
 	WLog_DBG(TAG, "---------------- End Dumping thread handles -------------");
+#endif
 }
 #endif
-#endif
